# Author: David Kuda
# Creation Date: October 2022
# License: MIT License

# This module makes connecting to databases with SSH and
# sql alchemy conn_strings easier and is meant to be used
# in jupyter lab / notebooks with the fantastic ipython
# magic "sql" package.

import os

from .connection_string import get_conn_str
from .ipython_magic import invoke_ipython_sql_magic
from .settings import get_settings
from .ssh_tunnel import SSHTunnel


def connect(environment: str):
    """Build conn_string from .settings.toml, export env_var, open ssh tunnel (if given).

    conn_string will be made available as env var DATABASE_URL and DATABASE_URL_ + environment.

    Args:
        environment (string):
            An environment corresponding to the objects in ".settings.toml", e.g.
            "production" or "staging".

    Returns:
        None
    """
    if environment is None:
        environment = os.environ["ENVIRONMENT"]

    settings = get_settings(environment)

    if settings.ssh_cmd is not None:
        SSHTunnel(settings.ssh_cmd).main()

    invoke_ipython_sql_magic()

    conn_str = get_conn_str(settings)
    env_var = "DATABASE_URL_" + environment.upper().replace("-", "_")
    os.environ[env_var] = conn_str

    os.environ["DATABASE_URL"] = conn_str

    return


if __name__ == "ndbdsession.sqlconn":
    # this block would run on import
    pass



from .settings import Settings


def get_conn_str(settings: Settings) -> str:
    """Returns db connection string that can be used with sqlalchemy.

    Args:
        config (Settings): Config of db connection.

    Returns:
        A string that can be used to connect to the database.

        Example / Format: "postgresql://user:password@url:port/database"
    """
    s = settings

    if settings.ssh_cmd:
        url = "127.0.0.1"
    else:
        url = settings.db_url

    return f"postgresql://{s.user}:{s.password}@{url}:{s.port}/{s.database}"



import os
import re
import subprocess
from dataclasses import dataclass
from urllib import parse

import tomllib

from .exceptions import EnvironmentDoesNotExistError


@dataclass
class Settings:
    db_driver: str
    user: str
    password: str
    db_url: str
    port: str
    database: str
    ssh_cmd: str


def get_settings(environment: str) -> Settings:
    settings = parse_toml()

    if environment not in settings.keys():
        raise EnvironmentDoesNotExistError(
            "Invalid environment, no such config / settings."
        )

    s: dict = settings[environment]

    return Settings(
        db_driver=s["db_driver"],
        user=s["user"],
        password=s["password"],
        db_url=s["db_url"],
        port=s["port"],
        database=s["database"],
        ssh_cmd=s.get("ssh_cmd"),
    )


def parse_toml(file_path: str = None) -> dict:
    """Parse .settings.toml that should be at top level of this repo."""

    if file_path is None:
        file_path = os.environ.get("SETTINGS_FILE_PATH")

    if file_path is None:
        # default to toplevel .settings.toml
        ps = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"], capture_output=True
        )
        root_dir = ps.stdout.decode().strip("\n")
        file_path = os.path.join(root_dir, ".settings.toml")

    with open(file_path, "rb") as f:
        data: dict = tomllib.load(f)

        for section in data.keys():

            # First, replace all strings with ${ENV_VAR} with os.environ[ENV_VAR]
            for k, v in data[section].items():

                # skip if not string
                if not isinstance(v, str):
                    continue

                if re.search(r"\${[A-Z_\d]*}", v):
                    env_var_key = v[2:-1]
                    data[section][k] = os.environ[env_var_key]

            # Then, replace password with quoted password
            for k, v in data[section].items():
                if k == "password":
                    data[section][k] = parse.quote(v)

        return data



from IPython import get_ipython


def invoke_ipython_sql_magic() -> None:
    ipython = get_ipython()

    already_invoked = ipython.find_magic("sql")
    if already_invoked is None:
        ipython.magic("load_ext sql")
    else:
        return


def connect_to_db(conn_str: str):
    get_ipython().magic(f"sql {conn_str}")
    return




import os
import signal
import subprocess


class SSHTunnel:
    def __init__(self, ssh_cmd: str) -> None:
        self.ssh_cmd = ssh_cmd

    def main(self):
        existing_ssh_tunnels = self._get_running_ssh_pids()

        if len(existing_ssh_tunnels) >= 1:
            self._terminate_processes_by_pids(existing_ssh_tunnels)

        self.open_ssh_tunnel()

    def open_ssh_tunnel(self):
        exit_status = os.system(self.ssh_cmd + "&")
        if exit_status != 0:
            raise Exception("Error: could not open ssh tunnel.")
        else:
            print("SSH Connection established.")

    def _get_running_ssh_pids(self) -> list[str]:
        ps = subprocess.Popen(("ps", "-A"), stdout=subprocess.PIPE)
        output = subprocess.check_output(("grep", "ssh"), stdin=ps.stdout)
        ps.wait()

        s = output.decode("utf-8")

        pids = []
        for line in s.split("\n"):
            if self.ssh_cmd in line:
                pid = line[:5]
                pids.append(pid)

        return pids

    @staticmethod
    def _terminate_processes_by_pids(pids: list[str]):
        for pid in pids:
            print(f"Terminating existing ssh process with pid {pid}")
            os.kill(int(pid), signal.SIGTERM)


if __name__ == "__main__":
    pass



from distutils.core import setup
from pathlib import Path
from setuptools import find_packages


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


# nbdbsession := notebook_database_session
setup(
    name="nbdbsession",
    version="0.2.5",
    description="Config mgt to connect to databases from jupyter notebooks.",
    author="David Kuda",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://github.com/davidkuda/jupyter_database_io",
    long_description=long_description,
    long_description_content_type='text/markdown',
)
